# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
A collection of RGI server applications for common scenarios.
"""

try:
    from ._base import (
        Router,
        ProxyApp,
    )
except ImportError:
    from ._basepy import (
        Router,
        ProxyApp,
    )


__all__ = (
    'AllowedMethods',
    'MethodFilter',
    'Router',
    'ProxyApp',
)


_ALLOWED_METHODS = {'GET', 'PUT', 'POST', 'HEAD', 'DELETE'}


class AllowedMethods:
    __slots__ = ('methods',)

    def __init__(self, *methods):
        for m in methods:
            if m not in _ALLOWED_METHODS:
                raise ValueError('bad method: {!r}'.format(m))
        self.methods = methods

    def __repr__(self):
        return 'AllowedMethods({})'.format(
            ', '.join(repr(m) for m in self.methods)
        )

    def __call__(self, app):
        return MethodFilter(app, self)

    def isallowed(self, m):
        return m in self.methods


class MethodFilter:
    __slots__ = ('app', 'allowed_methods')

    def __init__(self, app, allowed_methods):
        if not callable(app):
            raise TypeError(
                'app not callable: {!r}'.format(app)
            )
        if type(allowed_methods) is not AllowedMethods:
            raise TypeError(
                'allowed_methods: need a {!r}; got a {!r}: {!r}'.format(
                    AllowedMethods, type(allowed_methods), allowed_methods
                )
            )
        self.app = app
        self.allowed_methods = allowed_methods

    def __call__(self, session, request, api):
        if self.allowed_methods.isallowed(request.method):
            return self.app(session, request, api)
        return (405, 'Method Not Allowed', {}, None)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu.applib` module.
"""

from unittest import TestCase
import os
from random import SystemRandom

from ..misc import mkreq
from ..misc import TempServer
from ..client import Client
from .. import applib


random = SystemRandom()


METHODS = ('GET', 'PUT', 'POST', 'HEAD', 'DELETE')
BAD_METHODS = [
    '',
    'TRACE',
    'OPTIONS',
    'CONNECT',
    'PATCH',
    'GOT',
    'POT',
    'PUSH',
    'HELL',
    'REPEAT',
]
BAD_METHODS.extend(m.lower() for m in METHODS)
BAD_METHODS = tuple(BAD_METHODS)


class TestAllowedMethods(TestCase):
    def test_init(self):
        # No methods:
        inst = applib.AllowedMethods()
        self.assertEqual(inst.methods, tuple())

        # 0 to 5 good methods in random order:
        for count in range(6):
            methods = tuple(random.sample(METHODS, count))
            inst = applib.AllowedMethods(*methods)
            self.assertEqual(inst.methods, methods)

        # 0 to 5 good methods plus 1 bad method, all in random order:
        for bad in BAD_METHODS:
            for count in range(6):
                methods = random.sample(METHODS, count)
                methods.append(bad)
                random.shuffle(methods)
                with self.assertRaises(ValueError) as cm:
                    applib.AllowedMethods(*methods)
                self.assertEqual(str(cm.exception),
                    'bad method: {!r}'.format(bad)
                )

    def test_repr(self):
        # Static value sanity check: no methods
        inst = applib.AllowedMethods()
        self.assertEqual(repr(inst), 'AllowedMethods()')

        # Static value sanity check: one or more methods:
        inst = applib.AllowedMethods('HEAD')
        self.assertEqual(repr(inst), "AllowedMethods('HEAD')")
        inst = applib.AllowedMethods('POST', 'DELETE')
        self.assertEqual(repr(inst), "AllowedMethods('POST', 'DELETE')")
        inst = applib.AllowedMethods('PUT', 'HEAD', 'GET')
        self.assertEqual(repr(inst), "AllowedMethods('PUT', 'HEAD', 'GET')")
        inst = applib.AllowedMethods('GET', 'DELETE', 'POST', 'HEAD')
        self.assertEqual(repr(inst),
            "AllowedMethods('GET', 'DELETE', 'POST', 'HEAD')"
        )
        inst = applib.AllowedMethods('GET', 'PUT', 'POST', 'HEAD', 'DELETE')
        self.assertEqual(repr(inst),
            "AllowedMethods('GET', 'PUT', 'POST', 'HEAD', 'DELETE')"
        )

        # 0 to 5 good methods in random order:
        for count in range(6):
            methods = tuple(random.sample(METHODS, count))
            inst = applib.AllowedMethods(*methods)
            self.assertEqual(repr(inst),
                '{}({})'.format(
                    inst.__class__.__name__,
                    ', '.join(repr(m) for m in methods)
                )
            )

    def test_call(self):
        def app(session, request, api):
            return (200, 'OK', {}, None)

        for count in range(6):
            methods = tuple(random.sample(METHODS, count))
            inst = applib.AllowedMethods(*methods)

            # app not callable:
            bad = 'my_app'
            with self.assertRaises(TypeError) as cm:
                inst(bad)
            self.assertEqual(str(cm.exception),
                'app not callable: {!r}'.format(bad)
            )

            # All good:
            method_filter = inst(app)
            self.assertIs(type(method_filter), applib.MethodFilter)
            self.assertIs(method_filter.app, app)
            self.assertIs(method_filter.allowed_methods, inst)

    def test_isallowed(self):
        inst = applib.AllowedMethods('POST')
        self.assertIs(inst.isallowed('POST'), True)
        self.assertIs(inst.isallowed('GET'), False)
        self.assertIs(inst.isallowed('PUT'), False)
        self.assertIs(inst.isallowed('HEAD'), False)
        self.assertIs(inst.isallowed('DELETE'), False)

        inst = applib.AllowedMethods('PUT', 'HEAD')
        self.assertIs(inst.isallowed('HEAD'), True)
        self.assertIs(inst.isallowed('PUT'), True)
        self.assertIs(inst.isallowed('POST'), False)
        self.assertIs(inst.isallowed('GET'), False)
        self.assertIs(inst.isallowed('DELETE'), False)

        for count in range(6):
            methods = tuple(random.sample(METHODS, count))
            inst = applib.AllowedMethods(*methods)
            for m in BAD_METHODS:
                result = (True if m in methods else False)
                self.assertIs(inst.isallowed(m), result)


class TestMethodFilter(TestCase):
    def test_init(self):
        def app(session, request, api):
            return (200, 'OK', {}, None)

        allowed_methods = applib.AllowedMethods('GET', 'HEAD')

        # app not callable:
        bad = 'my_app'
        with self.assertRaises(TypeError) as cm:
            applib.MethodFilter(bad, allowed_methods)
        self.assertEqual(str(cm.exception),
            'app not callable: {!r}'.format(bad)
        )

        # allowed_methods isn't an AllowedMethods instance:
        bad = frozenset(['GET', 'HEAD'])
        with self.assertRaises(TypeError) as cm:
            applib.MethodFilter(app, bad)
        self.assertEqual(str(cm.exception),
            'allowed_methods: need a {!r}; got a {!r}: {!r}'.format(
                applib.AllowedMethods, type(bad), bad
            )
        )

        # All good:
        inst = applib.MethodFilter(app, allowed_methods)
        self.assertIs(type(inst), applib.MethodFilter)
        self.assertIs(inst.app, app)
        self.assertIs(inst.allowed_methods, allowed_methods)

    def test_call(self):
        class App:
            def __init__(self, marker):
                self.__marker = marker

            def __call__(self, session, request, api):
                return (200, 'OK', {}, self.__marker)

        marker = os.urandom(16)
        app = App(marker)

        # No methods allowed:
        allowed_methods = applib.AllowedMethods()
        inst = applib.MethodFilter(app, allowed_methods)
        for m in METHODS:
            self.assertEqual(inst(None, mkreq(m, '/'), None),
                (405, 'Method Not Allowed', {}, None)
            )

        # One method allowed:
        for allowed in METHODS:
            allowed_methods = applib.AllowedMethods(allowed)
            inst = applib.MethodFilter(app, allowed_methods)
            for m in METHODS:
                request = mkreq(m, '/')
                response = inst(None, request, None)
                if m == allowed:
                    self.assertEqual(response,
                        (200, 'OK', {}, marker)
                    )
                else:
                    self.assertEqual(response,
                        (405, 'Method Not Allowed', {}, None)
                    )

        # All *good* methods allowed:
        good = list(METHODS)
        random.shuffle(good)
        allowed_methods = applib.AllowedMethods(*good)
        inst = applib.MethodFilter(app, allowed_methods)
        for m in METHODS:
            self.assertEqual(inst(None, mkreq(m, '/'), None),
                (200, 'OK', {}, marker)
            )


class TestProxyApp(TestCase):
    def test_live(self):
        class Endpoint:
            def __init__(self, marker):
                self.marker = marker

            def __call__(self, session, request, api):
                return (200, 'OK', {}, self.marker)

        marker = os.urandom(16)
        app1 = Endpoint(marker)
        server1 = TempServer(('127.0.0.1', 0), app1)
        client1 = Client(server1.address)

        app2 = applib.ProxyApp(client1)
        server2 = TempServer(('127.0.0.1', 0), app2)
        client2 = Client(server2.address)

        conn = client2.connect()
        r = conn.get('/', {})
        self.assertEqual(r.status, 200)
        self.assertEqual(r.reason, 'OK')
        self.assertEqual(r.headers, {'content-length': 16})
        self.assertIs(r.body.chunked, False)
        self.assertEqual(r.body.read(), marker)
        conn.close()




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Common HTTP parser and IO abstractions used by server and client.
"""

try:
    from ._base import (
        EmptyPreambleError,
        API,      APIType,
        Response, ResponseType,
        Range,
        ContentRange,
        Request,
        api,
        Connection,
        Session,
        _handle_requests,
    )
except ImportError:
    from ._basepy import (
        EmptyPreambleError,
        API,      APIType,
        Response, ResponseType,
        Range,
        ContentRange,
        Request,
        api,
        Connection,
        Session,
        _handle_requests,
    )


# FIXME: for compatibility with Degu < 0.16:
bodies = api


__all__ = (
    'EmptyPreambleError',
    'API', 'APIType',
    'Response', 'ResponseType',
    'Range',
    'ContentRange',
    'Request',
    'api',
    'Connection',
    'Session',
    '_handle_requests',
)

# Provide very clear TypeError messages:
_TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests the `degu.misc` module.
"""

from unittest import TestCase
from os import path

from degu.server import Request
from degu import misc


class TestFunctions(TestCase):
    def test_mkreq(self):
        r = misc.mkreq('GET', '/')
        self.assertIs(type(r), Request)
        self.assertEqual(r.method, 'GET')
        self.assertEqual(r.uri, '/')
        self.assertEqual(r.headers, {})
        self.assertIsNone(r.body)
        self.assertEqual(r.mount, [])
        self.assertEqual(r.path, [])
        self.assertIsNone(r.query)

        r = misc.mkreq('GET', '/?')
        self.assertIs(type(r), Request)
        self.assertEqual(r.method, 'GET')
        self.assertEqual(r.uri, '/?')
        self.assertEqual(r.headers, {})
        self.assertIsNone(r.body)
        self.assertEqual(r.mount, [])
        self.assertEqual(r.path, [])
        self.assertEqual(r.query, '')

        r = misc.mkreq('GET', '/?key=val')
        self.assertIs(type(r), Request)
        self.assertEqual(r.method, 'GET')
        self.assertEqual(r.uri, '/?key=val')
        self.assertEqual(r.headers, {})
        self.assertIsNone(r.body)
        self.assertEqual(r.mount, [])
        self.assertEqual(r.path, [])
        self.assertEqual(r.query, 'key=val')

        r = misc.mkreq('GET', '/foo/bar')
        self.assertIs(type(r), Request)
        self.assertEqual(r.method, 'GET')
        self.assertEqual(r.uri, '/foo/bar')
        self.assertEqual(r.headers, {})
        self.assertIsNone(r.body)
        self.assertEqual(r.mount, [])
        self.assertEqual(r.path, ['foo', 'bar'])
        self.assertIsNone(r.query)

        # shift=1:
        r = misc.mkreq('GET', '/foo/bar', shift=1)
        self.assertIs(type(r), Request)
        self.assertEqual(r.method, 'GET')
        self.assertEqual(r.uri, '/foo/bar')
        self.assertEqual(r.headers, {})
        self.assertIsNone(r.body)
        self.assertEqual(r.mount, ['foo'])
        self.assertEqual(r.path, ['bar'])
        self.assertIsNone(r.query)

        # shift=2:
        r = misc.mkreq('GET', '/foo/bar', shift=2)
        self.assertIs(type(r), Request)
        self.assertEqual(r.method, 'GET')
        self.assertEqual(r.uri, '/foo/bar')
        self.assertEqual(r.headers, {})
        self.assertIsNone(r.body)
        self.assertEqual(r.mount, ['foo', 'bar'])
        self.assertEqual(r.path, [])
        self.assertIsNone(r.query)

        # With headers:
        headers = {'content-length': 32}
        r = misc.mkreq('GET', '/', headers)
        self.assertIs(type(r), Request)
        self.assertEqual(r.method, 'GET')
        self.assertEqual(r.uri, '/')
        self.assertIs(r.headers, headers)
        self.assertEqual(r.headers, {'content-length': 32})
        self.assertIsNone(r.body)
        self.assertEqual(r.mount, [])
        self.assertEqual(r.path, [])
        self.assertIsNone(r.query)

        # With body:
        body = 'my body'
        r = misc.mkreq('GET', '/', headers, body)
        self.assertIs(type(r), Request)
        self.assertEqual(r.method, 'GET')
        self.assertEqual(r.uri, '/')
        self.assertIs(r.headers, headers)
        self.assertEqual(r.headers, {'content-length': 32})
        self.assertIs(r.body, body)
        self.assertEqual(r.mount, [])
        self.assertEqual(r.path, [])
        self.assertIsNone(r.query)

    def test_mkuri(self):
        self.assertEqual(misc.mkuri(), '/')
        self.assertEqual(misc.mkuri(query=None), '/')
        self.assertEqual(misc.mkuri(query=''), '/?')
        self.assertEqual(misc.mkuri(query='hello'), '/?hello')
        self.assertEqual(misc.mkuri(query='k=v'), '/?k=v')

        self.assertEqual(misc.mkuri('foo'), '/foo')
        self.assertEqual(misc.mkuri('foo', query=None), '/foo')
        self.assertEqual(misc.mkuri('foo', query=''), '/foo?')
        self.assertEqual(misc.mkuri('foo', query='hello'), '/foo?hello')
        self.assertEqual(misc.mkuri('foo', query='k=v'), '/foo?k=v')

        self.assertEqual(misc.mkuri('foo', ''), '/foo/')
        self.assertEqual(misc.mkuri('foo', '', query=None), '/foo/')
        self.assertEqual(misc.mkuri('foo', '', query=''), '/foo/?')
        self.assertEqual(misc.mkuri('foo', '', query='hello'), '/foo/?hello')
        self.assertEqual(misc.mkuri('foo', '', query='k=v'), '/foo/?k=v')

        self.assertEqual(misc.mkuri('foo', 'bar'), '/foo/bar')
        self.assertEqual(misc.mkuri('foo', 'bar', query=None), '/foo/bar')
        self.assertEqual(misc.mkuri('foo', 'bar', query=''), '/foo/bar?')
        self.assertEqual(misc.mkuri('foo', 'bar', query='hello'),
            '/foo/bar?hello'
        )
        self.assertEqual(misc.mkuri('foo', 'bar', query='k=v'), '/foo/bar?k=v')

        self.assertEqual(misc.mkuri('foo', 'bar', ''), '/foo/bar/')
        self.assertEqual(misc.mkuri('foo', 'bar', '', query=None), '/foo/bar/')
        self.assertEqual(misc.mkuri('foo', 'bar', '', query=''), '/foo/bar/?')
        self.assertEqual(misc.mkuri('foo', 'bar', '', query='hello'),
            '/foo/bar/?hello'
        )
        self.assertEqual(misc.mkuri('foo', 'bar', '', query='k=v'),
            '/foo/bar/?k=v'
        )

    def test_format_headers(self):
        self.assertEqual(misc.format_headers({}), b'')
        self.assertEqual(misc.format_headers({'foo': 'bar'}), b'foo: bar')
        self.assertEqual(
            misc.format_headers({'foo': 'bar', 'bar': 'baz'}),
            b'bar: baz\r\nfoo: bar'
        )

    def test_format_request(self):
        self.assertEqual(
            misc.format_request('GET', '/', {}),
            b'GET / HTTP/1.1'
        )
        self.assertEqual(
            misc.format_request('GET', '/', {'foo': 'bar'}),
            b'GET / HTTP/1.1\r\nfoo: bar'
        )
        self.assertEqual(
            misc.format_request('GET', '/', {'foo': 'bar', 'bar': 'baz'}),
            b'GET / HTTP/1.1\r\nbar: baz\r\nfoo: bar'
        )

    def test_format_response(self):
        self.assertEqual(
            misc.format_response(200, 'OK', {}),
            b'HTTP/1.1 200 OK'
        )
        self.assertEqual(
            misc.format_response(200, 'OK', {'foo': 'bar'}),
            b'HTTP/1.1 200 OK\r\nfoo: bar'
        )
        self.assertEqual(
            misc.format_response(200, 'OK', {'foo': 'bar', 'bar': 'baz'}),
            b'HTTP/1.1 200 OK\r\nbar: baz\r\nfoo: bar'
        )


class TestTempPKI(TestCase):
    def test_init(self):
        # Test when client_pki=True:
        pki = misc.TempPKI()
        self.assertGreater(path.getsize(pki.path(pki.server_ca_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_ca_id, 'ca')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_id, 'cert')), 0)
        self.assertGreater(path.getsize(pki.path(pki.client_ca_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.client_ca_id, 'ca')), 0)
        self.assertGreater(path.getsize(pki.path(pki.client_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.client_id, 'cert')), 0)

        # pki.server_sslconfig property:
        self.assertEqual(pki.server_sslconfig, {
            'cert_file': pki.path(pki.server_id, 'cert'),
            'key_file': pki.path(pki.server_id, 'key'),
            'ca_file': pki.path(pki.client_ca_id, 'ca'),
        })

        self.assertEqual(pki.anonymous_server_sslconfig, {
            'cert_file': pki.path(pki.server_id, 'cert'),
            'key_file': pki.path(pki.server_id, 'key'),
            'allow_unauthenticated_clients': True,
        })

        self.assertEqual(pki.client_sslconfig, {
            'ca_file': pki.path(pki.server_ca_id, 'ca'),
            'check_hostname': False,
            'cert_file': pki.path(pki.client_id, 'cert'),
            'key_file': pki.path(pki.client_id, 'key'),
        })

        self.assertEqual(pki.anonymous_client_sslconfig, {
            'ca_file': pki.path(pki.server_ca_id, 'ca'),
            'check_hostname': False,
        })

        self.assertTrue(path.isdir(pki.ssldir))
        pki.__del__()
        self.assertFalse(path.exists(pki.ssldir))
        pki.__del__()

        # Test when client_pki=False:
        pki = misc.TempPKI(client_pki=False)
        self.assertGreater(path.getsize(pki.path(pki.server_ca_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_ca_id, 'ca')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_id, 'key')), 0)
        self.assertGreater(path.getsize(pki.path(pki.server_id, 'cert')), 0)

        self.assertEqual(pki.anonymous_server_sslconfig, {
            'cert_file': pki.path(pki.server_id, 'cert'),
            'key_file': pki.path(pki.server_id, 'key'),
            'allow_unauthenticated_clients': True,
        })

        self.assertEqual(pki.anonymous_client_sslconfig, {
            'ca_file': pki.path(pki.server_ca_id, 'ca'),
            'check_hostname': False,
        })

        self.assertTrue(path.isdir(pki.ssldir))
        pki.__del__()
        self.assertFalse(path.exists(pki.ssldir))
        pki.__del__()




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit test helpers.
"""

import io
import os
from os import path
import tempfile
import shutil
from random import SystemRandom
import string

from degu import tables
from degu.sslhelpers import random_id


MAX_IO_SIZE = 16777216  # 16 MiB
random = SystemRandom()


def random_data():
    """
    Return random bytes between 1 and 34969 (inclusive) bytes long.

    In unit tests, this is used to simulate a random request or response body,
    or a random chunk in a chuck-encoded request or response body.
    """
    size = random.randint(1, 34969)
    return os.urandom(size)


def random_chunks():
    """
    Return between 0 and 10 random chunks (inclusive).

    There will always be 1 additional, final chunk, an empty ``b''``, as per the
    HTTP/1.1 specification.
    """
    count = random.randint(0, 10)
    chunks = [random_data() for i in range(count)]
    chunks.append(b'')
    return chunks


def random_identifier():
    return ''.join(random.choice(string.ascii_lowercase) for i in range(17))


def random_chunk_ext():
    if random.randrange(3) == 0:
        return None
    return (random_identifier(), random_identifier())

def random_chunk(size=None):
    if size is None:
        size = random.randint(1, 34969)
    assert type(size) is int and 0 <= size <= MAX_IO_SIZE
    ext = random_chunk_ext()
    data = os.urandom(size)
    return (ext, data)

def random_chunks2(count=None):
    """
    Return between 0 and 10 random chunks (inclusive).

    There will always be 1 additional, final chunk, an empty ``b''``, as per the
    HTTP/1.1 specification.
    """
    if count is None:
        count = random.randrange(10) + 1
    assert type(count) is int and count > 0
    chunks = [
        random_chunk() for i in range(count - 1)
    ]
    chunks.append(random_chunk(0))
    assert len(chunks) == count
    return tuple(chunks)


def iter_good(good, allowed):
    for i in range(len(good)):
        for g in allowed:
            also_good = bytearray(good)
            also_good[i] = g
            yield bytes(also_good)


def iter_bad(good, allowed):
    assert isinstance(good, bytes)
    assert isinstance(allowed, bytes)
    not_allowed = tables.invert(allowed)
    for i in range(len(good)):
        for b in not_allowed:
            bad = bytearray(good)
            bad[i] = b
            yield bytes(bad)


class TempDir:
    def __init__(self, prefix='unittest.'):
        self.dir = tempfile.mkdtemp(prefix=prefix)

    def __del__(self):
        shutil.rmtree(self.dir)

    def join(self, *parts):
        return path.join(self.dir, *parts)

    def mkdir(self, *parts):
        dirname = self.join(*parts)
        os.mkdir(dirname)
        return dirname

    def makedirs(self, *parts):
        dirname = self.join(*parts)
        os.makedirs(dirname)
        return dirname

    def touch(self, *parts):
        filename = self.join(*parts)
        open(filename, 'xb').close()
        return filename

    def create(self, *parts):
        filename = self.join(*parts)
        return (filename, open(filename, 'xb'))

    def write(self, data, *parts):
        (filename, fp) = self.create(*parts)
        fp.write(data)
        fp.close()
        return filename

    def prepare(self, content):
        filename = self.write(content, random_id())
        return open(filename, 'rb')


class DummySocket:
    __slots__ = ('_calls',)

    def __init__(self):
        self._calls = []

    def shutdown(self, how):
        self._calls.append(('shutdown', how))

    def recv_into(self, buf):
        self._calls.append(('recv_into', buf))

    def send(self, buf):
        self._calls.append(('send', buf))


class DummyFile:
    def __init__(self):
        self._calls = []

    def close(self):
        self._calls.append('close')


class MockSocket:
    __slots__ = ('_rfile', '_wfile', '_rcvbuf', '_recv_into_calls', '_calls')

    def __init__(self, data, rcvbuf=None):
        assert rcvbuf is None or (isinstance(rcvbuf, int) and rcvbuf > 0)
        self._rfile = io.BytesIO(data)
        self._wfile = io.BytesIO()
        self._rcvbuf = rcvbuf
        self._recv_into_calls = 0
        self._calls = []

    def recv_into(self, buf):
        assert isinstance(buf, memoryview)
        self._calls.append(('recv_into', len(buf)))
        if self._rcvbuf is not None and len(buf) > self._rcvbuf:
            buf = buf[0:self._rcvbuf]
        self._recv_into_calls += 1
        return self._rfile.readinto(buf)

    def send(self, data):
        return self._wfile.write(data)


class MockSocket2:
    __slots__ = ('__rfile', '__rcvbuf')

    def __init__(self, rfile, rcvbuf=None):
        assert rcvbuf is None or (type(rcvbuf) is int and rcvbuf > 0)
        self.__rfile = rfile
        self.__rcvbuf = rcvbuf

    def recv_into(self, buf):
        assert isinstance(buf, memoryview)
        if self.__rcvbuf is not None:
            buf = buf[0:self.__rcvbuf]
        return self.__rfile.readinto(buf)


class MockBodies:
    def __init__(self, **kw):
        for (key, value) in kw.items():
            assert key in ('Body', 'BodyIter', 'ChunkedBody', 'ChunkedBodyIter')
            setattr(self, key, value)


def iter_bodies_with_missing_object():
    names = ('Body', 'BodyIter', 'ChunkedBody', 'ChunkedBodyIter')

    def dummy_body():
        pass

    for name in names:
        kw = dict((key, dummy_body) for key in names)
        del kw[name]
        yield (MockBodies(**kw), name)


def iter_bodies_with_non_callable_object():
    names = ('Body', 'BodyIter', 'ChunkedBody', 'ChunkedBodyIter')

    def dummy_body():
        pass

    for name in names:
        kw = dict((key, dummy_body) for key in names)
        attr = random_identifier()
        kw[name] = attr
        yield (MockBodies(**kw), name, attr)


def build_uri(path, query):
    uri = '/' + '/'.join(path)
    if query is None:
        return uri
    return '?'.join([uri, query])


def iter_random_path():
    yield []
    for length in range(1, 5):
        p = [random_id() for i in range(length)]
        yield p
        yield p + [''] 


def iter_random_uri():
    queries = (
        None,
        random_id(),
        '{}={}'.format(random_id(), random_id()),
        '{}={}&{}={}'.format(random_id(), random_id(), random_id(), random_id())
    )
    for p in iter_random_path():
        for q in queries:
            yield (build_uri(p, q), p, q)


